// tmpOLEReadWrite.cpp : Defines the entry point for the application.
//

#include "stdafx.h"
#include <objbase.h>
#include <crtdbg.h>
#include <stdio.h>
#include <tchar.h>

#pragma warning(disable: 4786)

#include <vector>
using namespace std;




class OLEReadWrite
{
public:

    // a IStorage is the equivalent of a folder
    // a IStream is the equivalent of a file
    // The hierarchy of IStorage/IStream is stored within a single file

    HRESULT ReadAndDuplicate(LPCWSTR filename, LPCWSTR output_filename) {

        IStorage* src_stg = NULL;

	    //
	    // Open an OLE Storage Object
	    //
	    HRESULT hr = StgOpenStorage(filename, 
						    NULL, 
						    STGM_READ | STGM_SHARE_DENY_WRITE,
						    NULL, 
						    0, 
						    &src_stg);
        if (FAILED(hr))
            return hr;

        IStorage* dest_stg = NULL;

	    //
	    // Create an OLE Storage Object
	    //
        hr = StgCreateDocfile(output_filename, 
						STGM_READWRITE | STGM_CREATE | STGM_SHARE_EXCLUSIVE,
						0, 
						&dest_stg);

        if (FAILED(hr))
        {
            src_stg->Release();
            src_stg = NULL;
            return hr;
        }

        hr = ReadAndDuplicate(src_stg, dest_stg);

        dest_stg->Release();
        dest_stg = NULL;

        src_stg->Release();
        src_stg = NULL;

        return hr;
    }

    HRESULT ReadAndDuplicate(IStorage* src, IStorage* dest) {

        _ASSERT(src != NULL);
        _ASSERT(dest != NULL);

        if (!src || !dest)
            return E_POINTER;

	    IEnumSTATSTG* penum = NULL;

	    HRESULT hr = src->EnumElements(0, NULL, 0, &penum);

	    if ( FAILED(hr) )
		    return hr;

	    penum->Reset();

	    ULONG nb = 0;
	    STATSTG stg;
	    hr = penum->Next(1, &stg, &nb);

	    while (SUCCEEDED(hr) && nb > 0)
	    {
		    // is it a stream or a storage?
		    if (stg.type == STGTY_STORAGE)
		    {
                IStorage* src_subStg = NULL;
                IStorage* dest_subStg = NULL;

	            hr = src->OpenStorage(stg.pwcsName, 
						            NULL, 
						            STGM_READ | STGM_SHARE_EXCLUSIVE, 
						            NULL, 
						            0, 
						            &src_subStg);

                if (FAILED(hr))
                    return hr;

	            hr = dest->CreateStorage(stg.pwcsName, 
                                         STGM_READWRITE | STGM_DIRECT | STGM_SHARE_EXCLUSIVE, 
                                         0, 0, 
                                         &dest_subStg);
                if (FAILED(hr))
                {
                    src_subStg->Release();
                    src_subStg = NULL;
    
                    return hr;
                }

                hr = ReadAndDuplicate(src_subStg, dest_subStg);

                dest_subStg->Release();
                dest_subStg = NULL;

                src_subStg->Release();
                src_subStg = NULL;

                if (FAILED(hr))
                    return hr;

		    }
		    else if (stg.type == STGTY_STREAM)
		    {
	            IStream* src_strm = NULL;
	            IStream* dest_strm = NULL;

	            HRESULT hr = src->OpenStream(stg.pwcsName, 
                                             NULL, 
                                             STGM_READ | STGM_SHARE_EXCLUSIVE, 
                                             0, 
                                             &src_strm);
	            if ( FAILED(hr) )
		            return hr;

	            hr = dest->CreateStream(stg.pwcsName, 
                                        STGM_READWRITE | STGM_SHARE_EXCLUSIVE | STGM_DIRECT, 
                                        0, 0, 
                                        &dest_strm);
	            if ( FAILED(hr) )
	            {
		            src_strm->Release();
		            src_strm = NULL;
		            return hr;
	            }


	            // read the actual content of the stream
	            //

	            ULONG nBytesRead = 0;
	            ULONG nAmount = 0;

	            BYTE pBuffer[4096];

	            hr = src_strm->Read(pBuffer, 4096, &nBytesRead); // reads the first chunk

	            while ( SUCCEEDED(hr) && nBytesRead > 0 )
	            {
		            // write the chunk in the destination
		            {
			            ULONG nBytesWritten = 0;
			            ULONG cbWritten;
			            DWORD dwBufSize = nBytesRead;
			            int nbTries = 0;
			            while (dwBufSize > 0 && nbTries < 3)
			            {
				            hr = dest_strm->Write((VOID const HUGEP*)(pBuffer+nBytesWritten),dwBufSize, &cbWritten);
				            if (FAILED(hr))
				            {
					            nbTries++;
					            continue;
				            }

				            dwBufSize -= cbWritten;
				            nBytesWritten += cbWritten;
			            } // while all written bytes
		            }

		            // iterate
		            nBytesRead = 0;
		            hr = src_strm->Read(pBuffer, 4096, &nBytesRead); // reads the next chunk
	            }

	            dest_strm->Release();
	            dest_strm = NULL;

	            src_strm->Release();
	            src_strm = NULL;
			    
		    }

            if (FAILED(hr))
                break;

		    // iterate
		    nb = 0;
		    hr = penum->Next(1, &stg, &nb);
	    }

	    penum->Release();

        return hr;
    }

    HRESULT Close() {

    }

protected:

};




int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
 	// This sample code reads a vbaProject.bin or oleObjectxxx.bin file
    // and writes it to another OLE document. This scenario is meant
    // to show a generic read and a generic write of such file format.

    OLEReadWrite o;
    o.ReadAndDuplicate(L"vbaProject.bin", L"test.file");


	return 0;
}



